<?php

namespace Drupal\simple_sitemap_arbitrary_links;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;

/**
 * Provides form alterations for custom sitemap links.
 */
class CustomSitemapLinksFormAlter {

  /**
   * Alters the custom links form to add a table of editable links.
   *
   * @param array &$form
   *   The form array to modify, passed by reference.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form, providing access to form data and state.
   *
   * @throws \RuntimeException
   *   If database query or language retrieval fails.
   */
  public static function alterForm(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();

    // Fetch all existing custom sitemap links from the database, keyed by ID.
    try {
      $links = $db->select('custom_sitemap_links', 'c')
      // Include 'external' field.
        ->fields('c', ['id', 'url', 'priority', 'changefreq', 'lastmod', 'langcode', 'external'])
        ->orderBy('id')
        ->execute()
        ->fetchAllAssoc('id');
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to load custom sitemap links: ' . $e->getMessage());
    }

    // Add a fieldset to group custom link configuration fields.
    $form['custom_links'] = [
      '#type' => 'fieldset',
      '#title' => t('Custom Arbitrary Links'),
      '#tree' => TRUE,
      '#description' => t('Define custom URLs for the XML sitemap, independent of Drupal entities.'),
      '#prefix' => '<div id="custom-links-wrapper">',
      '#suffix' => '</div>',
    ];

    $form['custom_links']['links'] = [
      '#type' => 'table',
      '#header' => [
        t('External'),
        t('URL'),
        t('Priority'),
        t('Change Frequency'),
        t('Last Modified'),
        t('Language'),
        t('Operations'),
      ],
    ];

    // Retrieve available languages for the language dropdown.
    try {
      $languages = \Drupal::languageManager()->getLanguages();
      $lang_options = array_map(fn($lang) => $lang->getName(), $languages);
      $default_lang = \Drupal::languageManager()->getDefaultLanguage()->getId();
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to retrieve language options: ' . $e->getMessage());
    }

    // Populate the table with rows for each existing link.
    foreach ($links as $id => $link) {
      $form['custom_links']['links'][$id] = [
        'external' => [
          '#type' => 'checkbox',
          '#title' => t('External'),
          '#default_value' => $link->external ?? 0,
        ],
        'url' => [
          '#type' => 'textfield',
          '#default_value' => $link->url,
          '#size' => 40,
        ],
        'priority' => [
          '#type' => 'select',
            // Fallback to 1.0 if not set.
          '#default_value' => $link->priority ?? '1.0',
          '#options' => array_combine(
                    $p = ['0.1', '0.2', '0.3', '0.4', '0.5', '0.6', '0.7', '0.8', '0.9', '1.0'],
                    $p
          ),
        ],
        'changefreq' => [
          '#type' => 'select',
          '#default_value' => $link->changefreq ?? '',
          '#options' => ['' => t('- Not specified -')] + array_map(
                        't',
                        ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']
          ),
        ],
        'lastmod' => [
          '#type' => 'date',
            // Convert timestamp to date input format.
          '#default_value' => $link->lastmod ? date('Y-m-d', strtotime($link->lastmod)) : '',
        ],
        'langcode' => [
          '#type' => 'select',
            // Use site default if unset.
          '#default_value' => $link->langcode ?? $default_lang,
          '#options' => $lang_options,
        ],
        'remove' => [
          '#type' => 'submit',
          '#value' => t('Remove'),
            // Unique name to identify this button.
          '#name' => "remove_$id",
        // Custom attribute to pass the link ID to the submit handler.
          '#delta' => $id,
          '#submit' => [[self::class, 'removeLink']],
        ],
      ];
    }

    // Add a button to trigger AJAX addition of a new link row.
    $form['custom_links']['add_link'] = [
      '#type' => 'submit',
      '#value' => t('Add new link'),
      '#ajax' => [
        'callback' => [self::class, 'addLinkAjax'],
    // Target for AJAX replacement.
        'wrapper' => 'custom-links-wrapper',
        'event' => 'click',
      ],
    ];

    // Attach the JS library for scrolling if triggered by a form action.
    if ($form_state->get('scroll_to_custom_links')) {
      $form['#attached']['library'][] = 'simple_sitemap_arbitrary_links/scroll_to_custom_links';
    }

    // Register the custom submit handler for saving form changes.
    $form['#submit'][] = [self::class, 'submitForm'];
  }

  /**
   * AJAX callback to add a new link to the database and rebuild the form.
   *
   * @param array &$form
   *   The form array, passed by reference.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   The AJAX response to update the form with a new row.
   *
   * @throws \RuntimeException
   *   If the database insert or form rebuild fails.
   */
  public static function addLinkAjax(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();
    // ISO 8601 format for current UTC time.
    $now = (new \DateTime('now', new \DateTimeZone('UTC')))->format('c');

    // Insert a new empty link with default values.
    try {
      $new_id = $db->insert('custom_sitemap_links')
        ->fields([
      // Placeholder URL, to be edited by the user.
          'url' => '',
          'priority' => '1.0',
          'changefreq' => '',
          'lastmod' => $now,
          'langcode' => \Drupal::languageManager()->getDefaultLanguage()->getId(),
          'external' => 0,
        ])
        ->execute();
      if (!$new_id) {
        throw new \RuntimeException('Failed to insert new sitemap link.');
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Error adding new sitemap link: ' . $e->getMessage());
    }

    // Mark the form for rebuilding to reflect the new link.
    $form_state->setRebuild(TRUE);

    // Rebuild the form to include the new link row.
    try {
      $rebuilt_form = \Drupal::formBuilder()
        ->rebuildForm(
                $form_state->getFormObject()->getFormId(),
                $form_state,
                $form
            );
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to rebuild form after adding link: ' . $e->getMessage());
    }

    // Return an AJAX response to replace the custom links section.
    return (new AjaxResponse())->addCommand(new ReplaceCommand('#custom-links-wrapper', $rebuilt_form['custom_links']));
  }

  /**
   * Submit handler to remove a link from the database.
   *
   * @param array &$form
   *   The form array, passed by reference.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form, containing the triggering element.
   *
   * @throws \RuntimeException
   *   If the link ID is invalid or deletion fails.
   */
  public static function removeLink(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();

    // Validate that the triggering element provides a numeric ID.
    if (!isset($trigger['#delta']) || !is_numeric($trigger['#delta'])) {
      throw new \RuntimeException('Invalid link ID for removal.');
    }
    $id = $trigger['#delta'];

    // Delete the specified link from the database.
    try {
      $deleted = \Drupal::database()->delete('custom_sitemap_links')
        ->condition('id', $id)
        ->execute();
      if ($deleted === 0) {
        throw new \RuntimeException("No link found with ID $id to remove.");
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to remove sitemap link: ' . $e->getMessage());
    }

    // Clear form values and input to reset the form state.
    $form_state->setValues([]);
    $form_state->setUserInput([]);

    // Trigger a scroll to the custom links section via JS on rebuild.
    $form_state->set('scroll_to_custom_links', TRUE);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Submit handler to save form changes and regenerate the sitemap.
   *
   * @param array &$form
   *   The form array, passed by reference.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form, containing submitted values.
   *
   * @throws \RuntimeException
   *   If database updates or sitemap regeneration fails.
   */
  public static function submitForm(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();
    $base_url = \Drupal::request()->getSchemeAndHttpHost();
    $values = $form_state->getValue(['custom_links', 'links']) ?: [];

    // Process each submitted link.
    foreach ($values as $id => $link) {
      if (!empty($link['url'])) {
        $url = $link['url'];

        // Skip URL normalization if the link is marked as external.
        if (empty($link['external'])) {
          // Normalize internal URLs: prepend a slash if relative,
          // or full base URL if it starts with a slash.
          $url = (strpos($url, '/') === 0 || preg_match('#^https?://#', $url)) ? $url : "/$url";
          $url = (strpos($url, '/') === 0) ? $base_url . $url : $url;
        }

        // Convert the last modified date to ISO 8601 format, or NULL if unset.
        try {
          $lastmod = !empty($link['lastmod']) ? (new \DateTime($link['lastmod'], new \DateTimeZone('UTC')))->format('c') : NULL;
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Invalid last modified date for link ID ' . $id . ': ' . $e->getMessage());
        }

        // Update or insert the link into the database using a merge query.
        try {
          $db->merge('custom_sitemap_links')
          // Cast ID to string to avoid type assertion issues.
            ->key('id', (string) $id)
            ->fields([
              'url' => $url,
              'priority' => $link['priority'],
              'changefreq' => $link['changefreq'],
              'lastmod' => $lastmod,
              'langcode' => $link['langcode'],
              'external' => $link['external'] ?? 0,
            ])
            ->execute();
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Failed to save sitemap link ID ' . $id . ': ' . $e->getMessage());
        }
      }
    }

    // Regenerate the sitemap if the user checked the "Regenerate now" option.
    if ($form_state->getValue('simple_sitemap_regenerate_now')) {
      try {
        \Drupal::service('simple_sitemap.generator')->generate();
      }
      catch (\Exception $e) {
        throw new \RuntimeException('Failed to regenerate sitemap: ' . $e->getMessage());
      }
    }
  }

}
